From ec57a937320f8cd646d24fb8a754631e7f550e75 Mon Sep 17 00:00:00 2001
From: Thomas Munro <thomas.munro@gmail.com>
Date: Tue, 4 Feb 2020 21:39:30 +1300
Subject: [PATCH 1/3] Fix shared typmod bug on DSM slot exhaustion.

If find_or_make_matching_shared_tupledesc() raises an error, an
entry with a null tupdesc could be left behind in RecordCacheHash,
causing a crash on a later lookup.

Fix by creating the entry after
find_or_make_matching_shared_tupledesc() returns.

Defect in commit cc5f8136.  Back-patch to 11.
---
 src/backend/utils/cache/typcache.c | 9 ++++++---
 1 file changed, 6 insertions(+), 3 deletions(-)

diff --git a/src/backend/utils/cache/typcache.c b/src/backend/utils/cache/typcache.c
index cdf6331a97..91db6b6008 100644
--- a/src/backend/utils/cache/typcache.c
+++ b/src/backend/utils/cache/typcache.c
@@ -1800,15 +1800,14 @@ assign_record_type_typmod(TupleDesc tupDesc)
 	/* Find or create a hashtable entry for this tuple descriptor */
 	recentry = (RecordCacheEntry *) hash_search(RecordCacheHash,
 												(void *) &tupDesc,
-												HASH_ENTER, &found);
-	if (found && recentry->tupdesc != NULL)
+												HASH_FIND, NULL);
+	if (recentry)
 	{
 		tupDesc->tdtypmod = recentry->tupdesc->tdtypmod;
 		return;
 	}
 
 	/* Not present, so need to manufacture an entry */
-	recentry->tupdesc = NULL;
 	oldcxt = MemoryContextSwitchTo(CacheMemoryContext);
 
 	/* Look in the SharedRecordTypmodRegistry, if attached */
@@ -1822,6 +1821,10 @@ assign_record_type_typmod(TupleDesc tupDesc)
 	}
 	ensure_record_cache_typmod_slot_exists(entDesc->tdtypmod);
 	RecordCacheArray[entDesc->tdtypmod] = entDesc;
+	recentry = (RecordCacheEntry *) hash_search(RecordCacheHash,
+												(void *) &tupDesc,
+												HASH_ENTER, &found);
+	Assert(!found);
 	recentry->tupdesc = entDesc;
 
 	/* Assign a unique tupdesc identifier, too. */
-- 
2.23.0

